from typing import Optional

import numpy as np
from scipy.integrate import simpson
from scipy.interpolate import interp1d

from .pattern import Pattern
from .methods import FourierTransformMethod


def calculate_sq(
    normalized_pattern: Pattern,
    f_squared_mean: np.ndarray,
    f_mean_squared: np.ndarray,
) -> Pattern:
    """
    Calculates the Faber Ziman structure factor, using the equation:
    S(Q) = (n * Intensity - incoherent_scattering - <f>^2-)/<f^2> + 1
    where n is the normalization factor and f are the scattering factors.
    The function takes in the already normalized intensity (incoherent scattering is also subtracted)

    :param normalized_pattern:    Pattern with q and (n * Intensity - incoherent scattering) as x and y
    :param f_squared_mean:        <f^2> - mean squared scattering factor for each q value in the pattern
    :param f_mean_squared:        <f>^2 - squared mean scattering factor for each q value in the pattern

    :return: S(Q) pattern
    """
    sq = (normalized_pattern.y - f_squared_mean + f_mean_squared) / f_mean_squared
    return Pattern(normalized_pattern.x, sq)


def calculate_fr(
    sq_pattern: Pattern,
    r: Optional[np.ndarray] = None,
    use_modification_fcn: bool = False,
    method: str = "integral",
) -> Pattern:
    """
    Calculates F(r) from a given S(Q) pattern for r values.
    If r is None, a range from 0 to 10 with step 0.01 is used.
    A Lorch modification function of the form:

        m = sin(q*pi/q_max)/(q*pi/q_max)

    can be used to address issues with a low q_max. This will broaden the sharp peaks in g(r)

    :param sq_pattern:              Structure factor S(Q) with lim_inf S(Q) = 1 and unit(q)=A^-1
    :param r:                       numpy array giving the r-values for which F(r) will be calculated,
                                    default is 0.01 to 10 with 0.01 as a step. units should be in Angstrom.
    :param use_modification_fcn:    boolean flag whether to use the Lorch modification function
    :param method:                  determines the method used for calculating fr, possible values are:
                                            - 'integral' solves the Fourier integral, by calculating the integral
                                            - 'fft' solves the Fourier integral by using fast fourier transformation

    :return: F(r) pattern
    """
    if r is None:
        r = np.arange(0.0, 10.005, 0.01)
        r[0] = 1e-10  # to avoid division by zero

    q, sq = sq_pattern.data
    if use_modification_fcn:
        modification = np.sin(q * np.pi / np.max(q)) / (q * np.pi / np.max(q))
    else:
        modification = 1

    if method == "integral" or method == FourierTransformMethod.INTEGRAL:
        fr = (
            2.0
            / np.pi
            * np.trapz(
                modification * q * (sq - 1) * np.array(np.sin(np.outer(q.T, r))).T, q
            )
        )
    elif method == "fft" or method == FourierTransformMethod.FFT:
        q_step = q[1] - q[0]
        r_step = r[1] - r[0]

        n_out = np.max([len(q), int(np.pi / (r_step * q_step))])
        n_out = 2 ** int(np.ceil(np.log2(n_out)))

        # find the number of q points needed to resolve the r-space
        q_max_target = 2 * np.pi / r_step
        n_target = int(np.ceil(q_max_target / q_step))

        # Round up to the next power of 2 for fastest possible fft
        n_out = 2 ** int(np.ceil(np.log2(n_target)))

        q_max_for_ifft = n_out * q_step

        f_q = modification * q * (sq - 1)
        y_for_ifft = np.zeros(n_out)
        y_for_ifft[: len(f_q)] = f_q

        ifft_result = np.fft.ifft(y_for_ifft) * 2 * q_max_for_ifft / np.pi
        ifft_imag = np.imag(ifft_result)[:n_out]
        ifft_x_step = 2 * np.pi / q_max_for_ifft
        ifft_x = np.arange(n_out) * ifft_x_step

        fr = np.interp(r, ifft_x, ifft_imag)
    else:
        raise NotImplementedError(
            "{} is not an allowed method for calculate_fr".format(method)
        )
    return Pattern(r, fr)


def calculate_sq_from_fr(
    fr_pattern: Pattern,
    q: Optional[np.ndarray] = None,
    method: str = "integral",
    use_modification_fcn: bool = False,
) -> Pattern:
    """
    Calculates S(Q) from an F(r) pattern for given q values.

    :param fr_pattern:              input F(r) pattern
    :param q:                       numpy array giving the q-values for which S(q) will be calculated,

    :return: F(r) pattern
    """
    if q is None:
        q = np.arange(0.0, 25.005, 0.05)
        q[0] = 1e-10  # to avoid division by zero

    r, fr = fr_pattern.data
    if method == "integral" or method == FourierTransformMethod.INTEGRAL:
        iq = np.trapz(fr * np.array(np.sin(np.outer(r.T, q))).T, r)
    elif method == "fft" or method == FourierTransformMethod.FFT:
        r_step = r[1] - r[0]
        q_step = q[1] - q[0]

        n_out = max(len(r), int(np.pi / (q_step * r_step)))
        n_out = 2 ** int(np.ceil(np.log2(n_out)))

        fr_padded = np.zeros(n_out)
        fr_padded[: len(fr)] = fr

        r_max_for_ifft = n_out * r_step

        fft_result = np.fft.ifft(fr_padded) * r_max_for_ifft
        fft_imag = np.imag(fft_result)[:n_out]
        fft_q_step = 2 * np.pi / (n_out * r_step)
        fft_q = np.arange(n_out) * fft_q_step

        iq = np.interp(q, fft_q, fft_imag)
    else:
        raise NotImplementedError(f"{method} is not a valid method for calculate_sq")

    if use_modification_fcn:
        modification = np.sin(q * np.pi / np.max(q)) / (q * np.pi / np.max(q))
    else:
        modification = 1

    if use_modification_fcn:
        # when using the modification function we get issues at large q values, since it is very close to 0 there
        # we need to extrapolate a couple of points to avoid issues

        eps = 1e-9

        # find area where there is no problem with the modification function
        valid = modification * q > eps
        iq_new = np.empty_like(q)
        iq_new[:] = np.nan  # initialize

        iq_new[valid] = iq[valid] / (q[valid] * modification[valid])

        # extrapolate the last few unstable points
        last_valid = np.where(valid)[0][-1]
        if last_valid < len(q) - 1:
            f = interp1d(
                q[valid], iq_new[valid], kind="linear", fill_value="extrapolate"
            )
            iq_new[~valid] = f(q[~valid])

        sq = 1 + iq_new
    else:
        sq = 1 + iq / q

    return Pattern(q, sq)


def calculate_sq_from_gr(
    gr_pattern: Pattern,
    q: np.ndarray,
    atomic_density: float,
    method: str = "integral",
) -> Pattern:
    """
    Performs a back Fourier transform from the pair distribution function g(r)

    :param gr_pattern:      g(r) pattern
    :param q:               numpy array of q values for which S(Q) should be calculated
    :param atomic_density:  number_density in atoms/A^3

    :return: S(Q) pattern
    """
    r, gr = gr_pattern.data

    # removing the nan value at the first index, which is caused by the division by zero when r started from zero
    if np.isnan(gr[0]):
        gr[0] = 0
    fr_pattern = Pattern(r, (gr - 1) * (4.0 * np.pi * r * atomic_density))
    return calculate_sq_from_fr(fr_pattern, q)


def calculate_gr(fr_pattern: Pattern, atomic_density: float) -> Pattern:
    """
    Calculates a g(r) pattern from a given F(r) pattern and the atomic density

    :param fr_pattern:     F(r) pattern
    :param atomic_density:  atomic density in atoms/A^3

    :return: g(r) pattern
    """
    r, f_r = fr_pattern.data
    r_safe = np.where(r == 0, 1e-10, r)
    g_r = 1 + f_r / (4.0 * np.pi * r_safe * atomic_density)
    return Pattern(r, g_r)



# -*- coding: utf-8 -*-
import re
from typing import Optional, Union, Dict
from collections import defaultdict
from copy import copy

import numpy as np
from scipy import interpolate
import scipy.constants as const

import lmfit

from .scattering_factors import (
    calculate_coherent_scattering_factor,
    calculate_incoherent_scattered_intensity,
)
from . import Pattern
from . import scattering_factors

__all__ = [
    "calculate_f_mean_squared",
    "calculate_f_squared_mean",
    "calculate_incoherent_scattering",
    "extrapolate_to_zero_linear",
    "extrapolate_to_zero_poly",
    "extrapolate_to_zero_spline",
    "calculate_s0",
    "extrapolate_to_zero_step",
    "convert_density_to_atoms_per_cubic_angstrom",
    "normalize_composition",
    "convert_two_theta_to_q_space",
    "convert_two_theta_to_q_space_raw",
    "calculate_weighting_factor",
]

Composition = Dict[str, Union[int, float]]


def parse_str_to_composition(formula: str) -> Composition:
    """
    Parses a chemical formula string into a dictionary with elements as keys and abundances as relative numbers.
    Typical examples are 'SiO2'-> {'Si': 1, 'O': 2} or 'Na2Si2O5' -> {'Na': 2, 'Si': 2, 'O': 5}

    :param formula: chemical formula string
    :return: Composition with elements as keys and abundances as relative numbers
    """
    # Remove all whitespace from the formula
    formula = re.sub(r"\s+", "", formula)

    def parse_segment(segment):
        element_pattern = r"([A-Z][a-z]*)(\d*\.\d+|\d*)"
        parsed = defaultdict(float)
        for element, count in re.findall(element_pattern, segment):
            parsed[element] += float(count) if count else 1.0
        return parsed

    def multiply_segment(segment_dict, multiplier):
        for element in segment_dict:
            segment_dict[element] *= multiplier
        return segment_dict

    def process_brackets(formula):
        bracket_patterns = [
            r"\(([^\(\)]+)\)(\d*\.\d+|\d*)",  # ()
            r"\[([^\[\]]+)\](\d*\.\d+|\d*)",  # []
            r"\{([^\{\}]+)\}(\d*\.\d+|\d*)",  # {}
        ]

        while any(re.search(pattern, formula) for pattern in bracket_patterns):
            for pattern in bracket_patterns:
                while re.search(pattern, formula):
                    matches = re.findall(pattern, formula)
                    for sub_formula, count in matches:
                        parsed_sub_formula = parse_segment(sub_formula)
                        multiplier = float(count) if count else 1.0
                        parsed_sub_formula = multiply_segment(
                            parsed_sub_formula, multiplier
                        )
                        sub_formula_string = "".join(
                            [
                                f"{el}{parsed_sub_formula[el]}"
                                for el in parsed_sub_formula
                            ]
                        )
                        formula = formula.replace(
                            f"({sub_formula}){count}", sub_formula_string, 1
                        )
                        formula = formula.replace(
                            f"[{sub_formula}]{count}", sub_formula_string, 1
                        )
                        formula = formula.replace(
                            f"{{{sub_formula}}}{count}", sub_formula_string, 1
                        )
        return formula

    # Process all brackets first
    formula = process_brackets(formula)

    # Parse the final expanded formula
    final_parsed = parse_segment(formula)
    return dict(final_parsed)


def calculate_f_mean_squared(
    composition: Composition, q: np.ndarray, sf_source="hajdu"
) -> np.ndarray:
    """
    Calculates the square of the mean form factor for a given composition over q.

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.
    """
    norm_elemental_abundances = normalize_composition(composition)

    res = np.zeros_like(q)
    for element, amount in norm_elemental_abundances.items():
        res = res + amount * calculate_coherent_scattering_factor(element, q, sf_source)

    return res**2


def calculate_f_squared_mean(
    composition: Composition, q: np.ndarray, sf_source: str = "hajdu"
) -> np.ndarray:
    """
    Calculates the mean of the squared form factors for a given composition for a given q vector.

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: mean of the squared form factors
    """
    norm_elemental_abundances = normalize_composition(composition)

    res = np.zeros_like(q)
    for key, value in norm_elemental_abundances.items():
        res = res + value * calculate_coherent_scattering_factor(key, q, sf_source) ** 2
    return res


def calculate_incoherent_scattering(
    composition: Composition, q: np.ndarray, sf_source: str = "hajdu"
) -> np.ndarray:
    """
    Calculates compton/incoherent scattering for a given composition

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: incoherent scattering array
    """
    norm_elemental_abundances = normalize_composition(composition)

    res = np.zeros_like(q)
    for key, value in norm_elemental_abundances.items():
        res += value * calculate_incoherent_scattered_intensity(key, q, sf_source)
    return res


def calculate_s0(composition: Composition, sf_source: str = "hajdu") -> float:
    """
    Calculates the I0 value for a given composition by extrapolating the coherent scattering factor to zero where I(Q)
    and the Compton scattering should have zero intensities.

    :param composition: dictionary with elements as keys and abundances as relative numbers
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: I0 value
    """
    f_mean_squared = calculate_f_mean_squared(composition, np.array([0]), sf_source)
    f_squared_mean = calculate_f_squared_mean(composition, np.array([0]), sf_source)

    s0 = -f_squared_mean / f_mean_squared + 1
    return s0[0]


def calculate_kn_correction(
    q: Union[np.ndarray, float], wavelength: float
) -> np.ndarray:
    """
    Calculates the Klein-Nishina correction for given q-values and wavelength. This
    is a correction which should be applied to the incoherent scattering intensity, due
    to the fact that the Compton scattering is chaning Energy with the scattering angle.

    :param q: Q value or numpy array with a unit of A^-1
    :param wavelength: wavelength in Angstrom

    :return: Klein-Nishina correction array for the given q-values
    """

    tth = 2 * np.arcsin(q * wavelength / (4 * np.pi))

    def P(wavelength, tth):
        return 1 / (
            1
            + (const.h / (wavelength * 1e-10 * const.m_e * const.c)) * (1 - np.cos(tth))
        )

    def KN_correction(p_values, tth):
        return (p_values**3 - p_values**2 * np.sin(tth) ** 2 + p_values) / (
            1 + np.cos(tth) ** 2
        )

    return KN_correction(P(wavelength, tth), tth)


def calculate_weighting_factor(
    composition: Composition,
    element_1: str,
    element_2: str,
    q: np.ndarray,
    sf_source="hajdu",
):
    """
    Calculates the weighting factor for an element-element contribution in a given composition (e.g. for Si-O in SiO2)

    :param composition: dictionary with elements as key string and abundances as relative numbers
    :param element_1: string giving element 1
    :param element_2: string giving element 2
    :param q: Q value or numpy array with a unit of A^-1
    :param sf_source: source of the scattering factors. Possible sources are 'hajdu' and 'brown_hubbell'.

    :return: weighting factor array
    """
    if element_1 == element_2:
        factor = 1
    else:
        factor = 2

    num_atoms = sum([val for _, val in composition.items()])
    f = {}  # form factors
    c = {}  # concentrations
    for element, val in composition.items():
        f[element] = calculate_coherent_scattering_factor(element, q, sf_source)
        c[element] = val / num_atoms

    f_sum_squared = np.zeros_like(q)
    for element, conc in c.items():
        f_sum_squared += f[element] * conc
    f_sum_squared = f_sum_squared**2

    return (
        factor
        * c[element_1]
        * c[element_2]
        * f[element_1]
        * f[element_2]
        / f_sum_squared
    )


def normalize_composition(composition: Composition) -> dict[str, float]:
    """
    normalizes elemental abundances to 1

    :param composition: dictionary with elements as key and abundances as relative numbers
    :return: normalized elemental abundances dictionary
    """
    sum = 0.0
    for key, val in composition.items():
        sum += val

    result = copy(composition)

    for key in result:
        result[key] /= sum

    return result


def convert_density_to_atoms_per_cubic_angstrom(
    composition: Composition, density: float
) -> float:
    """
    Converts densities given in g/cm3 into atoms per A^3

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param density: density in g/cm^3
    :return: density in atoms/A^3
    """

    # get_smallest abundance
    norm_elemental_abundances = normalize_composition(composition)
    mean_z = 0.0
    for element, concentration in norm_elemental_abundances.items():
        element = re.findall("[A-zA-Z]*", element)[0]
        mean_z += concentration * scattering_factors.atomic_weights["AW"][element]
    return density / mean_z * 0.602214129


def convert_density_to_grams_per_cubic_centimeter(
    composition: Composition, atomic_density: float
) -> float:
    """
    Converts densities given in atoms/A^3 into g/cm3

    :param composition: dictionary with elements as key and abundances as relative numbers
    :param atomic_density: density in atoms/A^3
    :return: density in g/cm^3
    """

    # get_smallest abundance
    norm_elemental_abundances = normalize_composition(composition)
    mean_z = 0.0
    for element, concentration in norm_elemental_abundances.items():
        element = re.findall("[A-zA-Z]*", element)[0]
        mean_z += concentration * scattering_factors.atomic_weights["AW"][element]
    return atomic_density * mean_z / 0.602214129


def extrapolate_to_zero_step(pattern: Pattern, y0: float = 0) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) by setting everything below the q_min of the pattern to y0 (default=0)

    :param pattern: input Pattern
    :param y0: y value at x = 0

    :return: extrapolated Pattern
    """
    x, y = pattern.data
    step = x[1] - x[0]
    low_x = np.arange(min(x) - step, 0 - step / 2, -step)[::-1]
    low_y = np.zeros(low_x.shape) + y0

    return Pattern(np.concatenate((low_x, x)), np.concatenate((low_y, y)))


def extrapolate_to_zero_linear(pattern: Pattern, y0: float = 0) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) using a linear function from the leftest point in the pattern

    :param pattern: input Pattern
    :param y0: y value at x = 0

    :return: new extrapolated Pattern (includes the original data)
    """
    x, y = pattern.data
    step = x[1] - x[0]
    low_x = np.arange(min(x) - step, 0 - step / 2, -step)[::-1]
    low_y = (y[0] - y0) / x[0] * low_x + y0
    return Pattern(np.concatenate((low_x, x)), np.concatenate((low_y, y)))


def extrapolate_to_zero_spline(
    pattern: Pattern,
    x_max: float,
    smooth_factor: Optional[float] = None,
    replace: bool = False,
    y0: float = 0,
) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) using a spline function.
    If the spline hits zero on the y-axis at an x value higher than 0 all values below this intersection
    will be set to zero

    :param pattern: input pattern
    :param x_max: defines the maximum x value within the spline will be fitted to the input pattern.
        this parameter should be larger than the minimum of the pattern x
    :param smooth_factor: defines the smoothing of the spline extrapolation please see numpy.UnivariateSpline manual for
        explanations
    :param replace: boolean flag whether to replace the data values in the fitted region (default = False)
    :param y0: y value at x = 0

    :return: extrapolated Pattern (includes the original one)
    """

    x, y = pattern.data
    x_step = x[1] - x[0]
    x_low = np.arange(min(x) - x_step, 0 - x_step / 2, -x_step)[::-1]

    x_overlap_ind = np.where(x < x_max)[0]

    x_intersection = np.concatenate(([0], x[x_overlap_ind]))
    y_intersection = np.concatenate(([y0], y[x_overlap_ind]))

    if replace:
        x_low = np.concatenate((x_low, x_intersection[1:]))
        ind = x > x_max
        x = x[ind]
        y = y[ind]

    spl = interpolate.UnivariateSpline(x_intersection, y_intersection, s=smooth_factor)
    y_low = spl(x_low)

    if type(y_low) is not np.ndarray:
        raise ValueError("Spline extrapolation failed")

    ind_below_zero = np.where(y_low < y0)[0]

    if len(ind_below_zero) > 0:
        y_low[: ind_below_zero[-1]] = y0

    return Pattern(np.concatenate((x_low, x)), np.concatenate((y_low, y)))


def extrapolate_to_zero_poly(
    pattern: Pattern, x_max: float, replace: bool = False, y0: float = 0
) -> Pattern:
    """
    Extrapolates a pattern to (0, y0) using a 2nd order polynomial:

    .. math::
        y = a*(x-c)+b*(x-c)^2 + y0


    if the polynomial extrapolation hits the value of y0 (default=0) at an x value higher than zero all y values below
    this intersection will be set to y0.

    :param pattern: input pattern
    :param x_max: defines the maximum x value within the polynomial will be fit
    :param replace: boolean flag whether to replace the data values in the fitted region (default = False)
    :param y0: y value at x = 0

    :return: extrapolated Pattern
    """

    x, y = pattern.data
    x_step = x[1] - x[0]

    x_fit = x[x < x_max]
    y_fit = y[x < x_max]

    params = lmfit.Parameters()
    params.add("a", value=1, min=0)
    params.add("b", value=1, min=0)
    params.add("c", value=1)

    def optimization_fcn(params):
        a = params["a"].value
        b = params["b"].value
        c = params["c"].value

        return y_fit - (x_fit - c) * a - (x_fit - c) ** 2 * b + y0

    result = lmfit.minimize(optimization_fcn, params)
    if not hasattr(result, "params") or result.params is None:
        raise ValueError("Polynomial extrapolation failed")

    a = result.params["a"].value
    b = result.params["b"].value
    c = result.params["c"].value

    x_low = np.arange(min(x), 0, -x_step)[::-1]

    if replace:
        x_low = np.concatenate((x_low, x_fit[1:]))
        ind = x > x_max
        x = x[ind]
        y = y[ind]

    y_low = a * (x_low - c) + b * (x_low - c) ** 2 - y0
    y_low[y_low < y0] = y0

    return Pattern(np.concatenate((x_low, x)), np.concatenate((y_low, y)))


def convert_two_theta_to_q_space_raw(
    two_theta: Union[float, np.ndarray], wavelength: float
) -> Union[float, np.ndarray]:
    """
    Converts two theta values into q space
    """
    return 4 * np.pi * np.sin(two_theta / 360.0 * np.pi) / wavelength


def convert_two_theta_to_q_space(pattern: Pattern, wavelength: float) -> Pattern:
    """
    Returns a new pattern with the x-axis converted from two theta into q space
    """
    q_pattern = copy(pattern)
    q_pattern.x = convert_two_theta_to_q_space_raw(q_pattern.x, wavelength)
    return q_pattern



import numpy as np

from .configuration import (
    Result,
    FitNormalization,
    IntNormalization,
    SampleConfig,
    DataConfig,
    CalculationConfig,
    Composition,
)
from .pattern import Pattern
from .methods import ExtrapolationMethod
from .normalization import normalize, normalize_fit
from .optimization import optimize_sq
from .transform import calculate_sq, calculate_fr, calculate_gr
from .utility import (
    calculate_f_squared_mean,
    calculate_f_mean_squared,
    calculate_incoherent_scattering,
    calculate_s0,
    calculate_kn_correction,
    extrapolate_to_zero_linear,
    extrapolate_to_zero_poly,
    extrapolate_to_zero_spline,
)

# only export functions inside of this file
__all__ = [
    "create_calculate_pdf_configs",
    "calculate_pdf",
    "validate_input",
]


def create_calculate_pdf_configs(
    data: Pattern,
    composition: Composition,
    density: float,
    bkg: Pattern = None,
    bkg_scaling: float = 1,
) -> tuple[DataConfig, CalculationConfig]:
    """
    Helper function to create a starting glassure input configuration.
    Automatically sets the q_min and q_max values to the first and last
    x-value of the data pattern - thus, the whole pattern gets transformed,
    when using this configuration.

    These two inputs can then be used with the *calculate_pdf* function in the calc module
    to calculate the structure factor S(q), the pair distribution function F(r) and
    the pair correlation function g(r).

    :param data: The data pattern.
    :param composition: The composition of the sample.
    :param density: The density of the sample in g/cm^3.
    :param bkg: The background pattern. None if no background is present.
    :param bkg_scaling: The scaling factor for the background pattern.

    :return: DataConfig, CalculationConfig
    """
    sample_config = SampleConfig(composition=composition, density=density)
    calculation_config = CalculationConfig(sample=sample_config)
    calculation_config.transform.q_min = data.x[0]
    calculation_config.transform.q_max = data.x[-1]

    data_config = DataConfig(data=data, bkg=bkg, bkg_scaling=bkg_scaling)
    return data_config, calculation_config


def calculate_pdf(
    data_config: DataConfig, calculation_config: CalculationConfig
) -> Result:
    """
    Process the input configuration and return the result.
    """
    validate_input(data_config, calculation_config)

    # create some shortcuts
    config = calculation_config
    transform = config.transform
    composition = config.sample.composition

    # subtract background
    if data_config.bkg is not None:
        sample = data_config.data - data_config.bkg * data_config.bkg_scaling
    else:
        sample = data_config.data

    # limit the pattern
    sample = sample.limit(transform.q_min, transform.q_max)

    # calculate form factor values
    q = sample.x
    f_squared_mean = calculate_f_squared_mean(
        composition, q, transform.scattering_factor_source
    )
    f_mean_squared = calculate_f_mean_squared(
        composition, q, transform.scattering_factor_source
    )
    incoherent_scattering = calculate_incoherent_scattering(
        composition, q, transform.scattering_factor_source
    )

    # klein-nishina correction
    if transform.kn_correction:
        if transform.wavelength is None:
            raise ValueError(
                "Wavelength must be set when using the Klein-Nishina correction."
            )
        inc_correction = calculate_kn_correction(q, transform.wavelength)
    else:
        inc_correction = 1

    # normalization
    if isinstance(transform.normalization, FitNormalization):
        opt = transform.normalization
        assert isinstance(opt, FitNormalization), (
            "Normalization config must be of type FitNormalizationConfig "
            + "when normalization method is set to 'fit'."
        )

        if opt.container_scattering is not None:
            container_scattering = (
                calculate_incoherent_scattering(
                    opt.container_scattering, q, transform.scattering_factor_source
                )
                * inc_correction
            )
        else:
            container_scattering = None

        norm_inc = (
            incoherent_scattering * inc_correction
            if opt.incoherent_scattering
            else None
        )

        params, norm = normalize_fit(
            sample_pattern=sample,
            f_squared_mean=f_squared_mean,
            incoherent_scattering=norm_inc,
            q_cutoff=opt.q_cutoff,
            method=opt.method,
            multiple_scattering=opt.multiple_scattering,
            container_scattering=container_scattering,
        )
    elif isinstance(transform.normalization, IntNormalization):
        opt = transform.normalization
        norm_inc = incoherent_scattering if opt.incoherent_scattering else None

        n, norm = normalize(
            sample_pattern=sample,
            atomic_density=config.sample.atomic_density,
            f_squared_mean=f_squared_mean,
            f_mean_squared=f_mean_squared,
            incoherent_scattering=norm_inc,
            attenuation_factor=opt.attenuation_factor,
        )

    else:
        raise NotImplementedError(
            "Only the FitNormalization and IntNormalizationnormalization methods are implemented at the moment."
        )

    # transform the pattern to S(Q)
    sq = calculate_sq(norm, f_squared_mean, f_mean_squared)

    # extrapolation
    if config.transform.extrapolation.s0 is not None:
        s0 = config.transform.extrapolation.s0
    else:
        s0 = calculate_s0(composition, transform.scattering_factor_source)

    extrapolation = transform.extrapolation
    match extrapolation.method:
        case ExtrapolationMethod.STEP:
            sq = sq.extend_to(0, s0)

        case ExtrapolationMethod.LINEAR:
            sq = extrapolate_to_zero_linear(sq, y0=s0)

        case ExtrapolationMethod.SPLINE:
            sq_cutoff = sq.x[0] + extrapolation.overlap
            sq = extrapolate_to_zero_spline(
                sq,
                x_max=sq_cutoff,
                y0=s0,
                replace=extrapolation.replace,
            )

        case ExtrapolationMethod.POLY:
            sq_cutoff = sq.x[0] + extrapolation.overlap
            sq = extrapolate_to_zero_poly(
                sq,
                x_max=sq_cutoff,
                y0=s0,
                replace=extrapolation.replace,
            )

        case _:
            raise NotImplementedError(
                f"Extrapolation method {extrapolation.method} not implemented."
            )

    # Kaplow optimization
    if config.optimize is not None:
        opt = config.optimize
        sq = optimize_sq(
            sq,
            atomic_density=config.sample.atomic_density,
            r_cutoff=opt.r_cutoff,
            r_step=transform.r_step,
            iterations=opt.iterations,
            use_modification_fcn=opt.use_modification_fcn,
            fourier_transform_method=transform.fourier_transform_method,
        )

    fr = calculate_fr(
        sq,
        use_modification_fcn=transform.use_modification_fcn,
        method=transform.fourier_transform_method,
        r=np.arange(
            transform.r_min,
            transform.r_max + transform.r_step * 0.5,
            transform.r_step,
        ),
    )

    gr = calculate_gr(
        fr,
        atomic_density=config.sample.atomic_density,
    )

    res = Result(
        calculation_config=config.model_copy(deep=True),
        sq=sq,
        fr=fr,
        gr=gr,
    )

    return res


def validate_input(data_config: DataConfig, calculation_config: CalculationConfig):
    """
    Validate the input configuration.
    """
    if data_config.data is None or not isinstance(data_config.data, Pattern):
        raise ValueError("Input data must be a Pattern object.")
    if data_config.bkg is not None and not isinstance(data_config.bkg, Pattern):
        raise ValueError("Background data must be a Pattern object.")

    if not calculation_config.sample.composition:  # empty composition dict
        raise ValueError("Composition must be set.")

    if not calculation_config.sample.atomic_density:
        raise ValueError("Atomic density must be set.")



from typing import Optional

import numpy as np
import lmfit

from .pattern import Pattern
from .scattering_factors import calculate_incoherent_scattered_intensity


def normalize(
    sample_pattern: Pattern,
    atomic_density: float,
    f_squared_mean: np.ndarray,
    f_mean_squared: np.ndarray,
    incoherent_scattering: Optional[np.ndarray] = None,
    attenuation_factor: float = 0.001,
) -> tuple[float, Pattern]:
    """
    Normalizes the sample data (already background subtracted and corrected) to
    atomic units using the Kroegh-Moe-Norman integral normalization. The normalization
    is performed for the the Faber-Ziman structure factor.

    :param sample_pattern:      background subtracted sample pattern
    :param atomic_density:      density in atoms per cubic Angstro
    :param f_squared_mean:      <f^2> - mean squared scattering factor
    :param f_mean_squared:      <f>^2 - squared mean scattering factor
    :param incoherent_scattering:
                                compton scattering from sample, if set to None, it will not be used
    :param attenuation_factor:  attenuation factor used in the exponential, in order to correct for the q cutoff

    :return:                    normalization factor, normalized Pattern (incoherent scattering already subtracted)
    """
    q, intensity = sample_pattern.data
    # calculate values for integrals
    if incoherent_scattering is None:
        incoherent_scattering = np.zeros_like(q)
    n1 = (
        q**2
        * (
            (f_squared_mean + incoherent_scattering)
            * np.exp(-attenuation_factor * q**2)
        )
        / f_mean_squared
    )
    n2 = q**2 * intensity * np.exp(-attenuation_factor * q**2) / f_mean_squared

    n = (-2 * np.pi**2 * atomic_density + np.trapz(n1, q)) / np.trapz(n2, q)

    return n, Pattern(q, n * intensity - incoherent_scattering)


def normalize_fit_lmfit(
    sample_pattern: Pattern,
    f_squared_mean: np.ndarray,
    incoherent_scattering: Optional[np.ndarray] = None,
    q_cutoff: float = 3,
    method: str = "squared",
    multiple_scattering: bool = False,
    container_scattering: Optional[np.ndarray] = None,
) -> tuple[lmfit.Parameters, Pattern]:
    """
    This function is deprecated and will be removed in the future. It is replaced by a new
    implementation that uses the least squares problem to solve for the normalization factor.
    The new implementation is more efficient and more accurate. In the simplest case with no
    multiple scattering or container scattering, it is 100 times faster than the old implementation.

    Estimates the normalization factor n for calculating S(Q) by fitting
        (Intensity*n-Multiple Scattering-Container Scattering*m)
    to
        (Incoherent Scattering + Self Scattering)
    where n, m and Multiple Scattering are free parameters. It uses lmfit to solve the least squares
    problem.

    :param sample_pattern:      background subtracted sample pattern with A^-1 as x unit
    :param incoherent_scattering:
                                compton scattering from sample, if set to None, it will not be used
                                array should contain values for each q-value of the sample pattern
    :param f_squared_mean:      <f^2> - mean squared scattering factor for each q-value of the
                                sample pattern
    :param q_cutoff:            q value above which the fitting will be performed, default = 4
    :param method:              specifies whether q^2 ("squared") or q (linear) should be used for
                                scaling the fit, this ensures that higher q values are weighted more
    :multiple_scattering:       flag whether multiple scattering should be included in the fit - the
                                current implementation is just to remove a constant value from the
                                input data
    :container_scattering:      extra scattering from the container, if set to None, it will not be used.
                                Example usecase is extra diamond compton scattering contribution, which
                                will increase with pressure in soller slit diamond anvil experiments.
                                The amount of this extra scattering contribution will be fitted and output
                                as a separate parameter n_container in the result. Length of the array should
                                be the same as the length of the sample pattern. Any corrections to this
                                scattering should be done before calling this function (e.g. Klein-Nishima
                                correction)

    :return:    lmfit parameter object with the fitted parameters (n, multiple, n_countainer),
                normalized Pattern (incoherent scattering already subtracted)

    """
    q, intensity = sample_pattern.data
    q_ind = np.where(q > q_cutoff)[0]

    q_cut = q[q_ind]
    intensity_cut = intensity[q_ind]

    assert len(q_cut) > 0, "No q values above the cutoff value"
    assert len(f_squared_mean) == len(
        q
    ), """f_squared_mean should have the same length as the
        sample pattern"""

    f_squared_mean_cut = f_squared_mean[q_ind]

    # calculate values for integrals
    if incoherent_scattering is None:
        incoherent_scattering = 0
        incoherent_scattering_cut = 0
    else:
        assert len(incoherent_scattering) == len(
            q
        ), """incoherent scattering should have the same length as the
        sample pattern"""

        incoherent_scattering_cut = incoherent_scattering[q_ind]

    if method == "squared":
        scaling = q_cut**2
    elif method == "linear":
        scaling = q_cut
    else:
        raise NotImplementedError(
            "{} is not an allowed method for fit_normalization_factor".format(method)
        )

    # prepare lmfit parameters
    params = lmfit.Parameters()
    params.add("n", value=1, min=0)

    if multiple_scattering:
        params.add("multiple", value=1, min=0)
    else:
        params.add("multiple", value=0, vary=False)

    if container_scattering is not None:
        assert len(container_scattering) == len(
            q
        ), """container scattering should have the same length as the sample pattern"""
        params.add("n_container", value=10, min=0)
        container_contribution = container_scattering
        container_contribution_cut = container_contribution[q_ind]
    else:
        params.add("n_container", value=0, vary=False)
        container_contribution = 0
        container_contribution_cut = 0

    def optimization_fcn(params):
        n = params["n"].value
        multiple = params["multiple"].value
        n_container = params["n_container"].value
        if container_scattering is not None:
            compton = (
                incoherent_scattering_cut + container_contribution_cut * n_container
            )
        else:
            compton = incoherent_scattering_cut

        theory = f_squared_mean_cut + compton
        return ((n * intensity_cut - multiple - theory) * scaling) ** 2

    out = lmfit.minimize(optimization_fcn, params)

    # prepare final output
    q_out = sample_pattern.x
    compton_out = (
        incoherent_scattering + container_contribution * out.params["n_container"].value
    )
    intensity_out = (
        out.params["n"].value * intensity - out.params["multiple"].value - compton_out
    )

    return out.params, Pattern(q_out, intensity_out)


def normalize_fit(
    sample_pattern: Pattern,
    f_squared_mean: np.ndarray,
    incoherent_scattering: Optional[np.ndarray] = None,
    q_cutoff: float = 3,
    method: str = "squared",
    multiple_scattering: bool = False,
    container_scattering: Optional[np.ndarray] = None,
) -> tuple[dict, Pattern]:
    """
    Estimates the normalization factor n for calculating S(Q) by solving the linear least squares problem
        (Intensity*n-Multiple Scattering-Container Scattering*m) = (Incoherent Scattering + Self Scattering)
    where n, m and Multiple Scattering are free parameters.
    The solution is found by solving the normal equations.

    :param sample_pattern:      background subtracted sample pattern with A^-1 as x unit
    :param f_squared_mean:      <f^2> - mean squared scattering factor for each q-value of the
                                sample pattern
    :param incoherent_scattering:
                                compton scattering from sample, if set to None, it will not be used
                                array should contain values for each q-value of the sample pattern
    :param q_cutoff:            q value above which the fitting will be performed, default = 4
    :param method:              specifies whether q^2 ("squared") or q (linear) should be used for
                                scaling the fit, this ensures that higher q values are weighted more
    :multiple_scattering:       flag whether multiple scattering should be included in the fit - the
                                current implementation is just to remove a constant value from the
                                input data
    :container_scattering:      extra scattering from the container, if set to None, it will not be used.
                                Example usecase is extra diamond compton scattering contribution, which
                                will increase with pressure in soller slit diamond anvil experiments.
                                The amount of this extra scattering contribution will be fitted and output
                                as a separate parameter n_container in the result. Length of the array should
                                be the same as the length of the sample pattern. Any corrections to this
                                scattering should be done before calling this function (e.g. Klein-Nishima
                                correction)

    :return:    dictionary with the fitted parameters (n, multiple, n_countainer),
                normalized Pattern (incoherent scattering already subtracted)
    """
    q, intensity = sample_pattern.data
    q_ind = np.where(q > q_cutoff)[0]

    q_cut = q[q_ind]
    intensity_cut = intensity[q_ind]
    f_squared_mean_cut = f_squared_mean[q_ind]

    assert len(q_cut) > 0, "No q values above the cutoff value"
    assert len(f_squared_mean) == len(q), "f_squared_mean must match q length"

    # Handle incoherent scattering
    if incoherent_scattering is None:
        incoherent_scattering = np.zeros_like(q)
        incoherent_cut = np.zeros_like(intensity_cut)
    else:
        assert len(incoherent_scattering) == len(q)
        incoherent_cut = incoherent_scattering[q_ind]

    # Set scaling weights
    if method == "squared":
        q_scaling = q_cut**2
    elif method == "linear":
        q_scaling = q_cut
    else:
        raise NotImplementedError(f"{method} is not a valid method")

    if not multiple_scattering and container_scattering is None:
        # If this is the case we can use a simpler solution, using the dot product
        # of the intensity and the f_squared_mean + incoherent scattering (which is
        # the objective function that we are fitting to)
        y = intensity_cut * q_scaling
        f = (f_squared_mean_cut + incoherent_cut) * q_scaling
        n = np.dot(y, f) / np.dot(y, y)

        result = {
            "n": n,
            "multiple": 0,
            "n_container": 0,
        }
        intensity_out = n * intensity - incoherent_scattering

        return result, Pattern(q, intensity_out)

    # If we have multiple scattering or container scattering, we need to use the lstsq solution
    # Build design matrix X (columns: intensity, container, constant)

    # Handle container scattering
    if container_scattering is not None:
        assert len(container_scattering) == len(q)
        container_cut = container_scattering[q_ind]
    else:
        container_cut = np.zeros_like(intensity_cut)

    # Compute scaling factors for normalization
    scale_intensity = np.max(np.abs(intensity_cut)) or 1.0
    scale_container = np.max(np.abs(container_cut)) or 1.0

    # Apply scaling to predictors
    X_cols = [
        intensity_cut / scale_intensity,  # for n
        container_cut / scale_container,  # for n_container
        np.ones_like(intensity_cut),  # for multiple (constant)
    ]
    if not multiple_scattering:
        X_cols[2] = np.zeros_like(intensity_cut)

    X = np.column_stack([col * q_scaling for col in X_cols])
    y = (f_squared_mean_cut + incoherent_cut) * q_scaling

    # Solve scaled least squares
    result = np.linalg.lstsq(X, y, rcond=None)
    coeffs = result[0]

    # Rescale coefficients back to physical values
    n = coeffs[0] / scale_intensity
    n_container = -coeffs[1] / scale_container
    multiple = -coeffs[2]
    # n_container and multiple are negative values, since the matrix lstsq solution
    # is assuming they are additive to the intensity (but of course we are later subtracting them).
    # This is why we need to negate them here.

    # Build normalized intensity
    compton_out = incoherent_scattering

    if container_scattering is not None:
        compton_out += container_scattering * n_container

    intensity_out = n * intensity - multiple - compton_out

    result = {
        "n": n,
        "n_container": n_container,
        "multiple": multiple,
    }

    return result, Pattern(q, intensity_out)



# -*- coding: utf-8 -*/:

from typing import Optional, Literal
from pydantic import BaseModel, Field, computed_field, field_validator

from .utility import (
    Composition,
    convert_density_to_atoms_per_cubic_angstrom,
    convert_density_to_grams_per_cubic_centimeter,
)
from .pattern import Pattern
from .methods import FourierTransformMethod, NormalizationMethod, ExtrapolationMethod
from .utility import parse_str_to_composition


class SampleConfig(BaseModel):
    composition: Composition = Field(
        default_factory=dict,
        description=(
            "Composition of the sample. Can be a dictionary of elements, "
            "e.g. {'Si': 1, 'O': 2}, and their abundances or a string of the form 'SiO2'."
        ),
    )
    density: Optional[float] = Field(
        default=None,
        description="Density in g/cm^3. Will be automatically updated when the atomic density is set",
    )

    @computed_field(
        description=(
            "The atomic density in atoms per cubic Angstrom. Will be "
            "automatically updated when density is set."
        )
    )
    @property
    def atomic_density(self) -> Optional[float]:
        if self.composition == {}:  # empty composition
            return None
        return convert_density_to_atoms_per_cubic_angstrom(
            self.composition, self.density
        )

    @atomic_density.setter
    def atomic_density(self, value: Optional[float]):
        if self.composition == {}:
            self.density = None
        else:
            self.density = convert_density_to_grams_per_cubic_centimeter(
                self.composition, value
            )

    @field_validator("composition", mode="before")
    @classmethod
    def _validate_composition(cls, v: Composition | str) -> Composition:
        if isinstance(v, str):
            return parse_str_to_composition(v)
        return v


class FitNormalization(BaseModel):
    TYPE: Literal["fit"] = Field(default="fit", description="Normalization type")
    q_cutoff: float = Field(
        default=3.0,
        description=(
            "Cutoff q in 1/A for the normalization. Only above this value "
            "the normalization is performed."
        ),
    )
    method: str = Field(
        default="squared",
        description=(
            "How to scale the values in respect to q during fitting. "
            '"linear" or "squared" are possible.'
        ),
    )
    multiple_scattering: bool = Field(
        default=False,
        description=(
            "Whether to consider multiple scattering - if true, the multiple "
            "scattering is approximated by a constant value."
        ),
    )
    incoherent_scattering: bool = Field(
        default=True,
        description=(
            "Whether to subtract the incoherent scattering during the normalization."
        ),
    )
    container_scattering: Optional[Composition] = Field(
        default=None,
        description=(
            "Composition of the container material in the experiment. Can be a dictionary of elements, "
            "e.g. {'Si': 1, 'O': 2}, and their abundances or a string of the form 'SiO2'. Only the "
            "incoherent scattering of the container is considered. The container "
            "scattering is subtracted from the total scattering and the amount is "
            "fitted by just muliplying it with a constant value. If None, no container scattering is considered."
        ),
    )

    @field_validator("container_scattering", mode="before")
    @classmethod
    def _validate_container_scattering(
        cls, v: Composition | str | None
    ) -> Optional[Composition]:
        if isinstance(v, str):
            return parse_str_to_composition(v)
        return v


class IntNormalization(BaseModel):
    TYPE: Literal["integral"] = Field(
        default="integral", description="Normalization type"
    )
    attenuation_factor: float = Field(
        default=0.001, description="Attenuation factor for the normalization"
    )
    incoherent_scattering: bool = Field(
        default=True,
        description="Whether to subtract the incoherent scattering during the normalization",
    )


class OptimizeConfig(BaseModel):
    r_cutoff: float = Field(
        default=1.4,
        description="Cutoff r for the Kaplow optimization scheme. Should be below the first peak in g(r).",
    )
    iterations: int = Field(
        default=5, description="Number of iterations for the Kaplow optimization."
    )
    use_modification_fcn: bool = Field(
        default=False,
        description="Whether to use the Lorch modification function during the optimization procedure. "
        + "This can be different from the transform configuration.",
    )


class ExtrapolationConfig(BaseModel):
    method: ExtrapolationMethod = Field(
        default=ExtrapolationMethod.STEP,
        description="Method for the extrapolation of the structure factor S(q) from q_min to zero.",
    )
    s0: Optional[float] = Field(
        default=None,
        description="Target value at S(0) for the extrapolation to. If is None, the theorethical value is used.",
    )
    overlap: float = Field(
        default=0.2,
        description="Overlap in q-space [1/A] for the extrapolation. E.g. the fitting range.",
    )
    replace: bool = Field(
        default=False,
        description="Whether to replace the original S(q) data in the overlap region with the extrapolated values.",
    )


class TransformConfig(BaseModel):
    q_min: float = Field(
        default=0.0,
        description=(
            "Minimum q in 1/Angstrom from the data. Below it will be "
            "extended to zero."
        ),
    )
    q_max: float = Field(
        default=10.0, description="Maximum q in 1/Angstrom from the data."
    )

    r_min: float = Field(
        default=0.0,
        description=(
            "Minimum r in Angstrom for the calculated pair distribution "
            "function g(r)."
        ),
    )
    r_max: float = Field(
        default=10.0,
        description=(
            "Maximum r in Angstrom for the calculated pair distribution "
            "function g(r)."
        ),
    )
    r_step: float = Field(
        default=0.01,
        description=(
            "Step size for the r values in Angstrom for the calculated pair "
            "distribution function g(r)."
        ),
    )

    normalization: FitNormalization | IntNormalization = Field(
        default_factory=FitNormalization,
        description=(
            "Normalization configuration model. Possible values are "
            ":class:`FitNormalization` or :class:`IntNormalization`."
        ),
    )

    extrapolation: ExtrapolationConfig = Field(
        default_factory=ExtrapolationConfig,
        description="Extrapolation configuration model.",
    )

    use_modification_fcn: bool = Field(
        default=False, description="Whether to use the Lorch modification function."
    )
    kn_correction: bool = Field(
        default=False,
        description=(
            "Whether to apply the Klein-Nishima correction to the Compton "
            "scattering of the sample and the container (defined in normalization)."
        ),
    )
    wavelength: Optional[float] = Field(
        default=None,
        description=(
            "Wavelength in Angstrom. Needs to be set for the "
            "Klein-Nishima correction."
        ),
    )

    fourier_transform_method: FourierTransformMethod = FourierTransformMethod.FFT

    scattering_factor_source: str = Field(
        default="brown_hubbell",
        description=(
            "Source of the scattering factors. Possible values are: 'brown_hubbell' "
            "for scattering factors from Brown et al., 2006 and Hubbell et al., 1975 "
            "or 'haijdu' for scattering factors from Hajdu et al. "
            "(Acta Cryst. (1992). A48, 344-352)."
        ),
    )


class CalculationConfig(BaseModel):
    """Main  calculation configuration model for the glassure data processing.
    Does not contain any data, but only the information how to process the dataset.

    To reuse the calculation config for a different calculation with some parameters changed, it is advised to use the
    model_copy(deep=True) method of the config.

    This will create a deep copy of the configuration object and not
    overwrite parameters of the original one. (see https://docs.pydantic.dev/latest/concepts/serialization/#model_copy
    for more information).

    For example:
    ```
    config = CalculationConfig()
    config.sample.composition = {"Si": Si, "O": 2}

    config_copy = config.model_copy(deep=True)
    config_copy.sample.composition = {"Ge": 1, "O": 2}
    ```
    """

    sample: SampleConfig = Field(
        default_factory=SampleConfig,
        description="Sample configuration model, containing the composition and density of the material.",
    )
    transform: TransformConfig = Field(
        default_factory=TransformConfig,
        description="""Transform configuration model, containing the normalization, transform and 
        extrapolation settings.""",
    )
    optimize: Optional[OptimizeConfig] = Field(
        default=None,
        description="Optimization configuration model. If None, no optimization is performed",
    )


class DataConfig(BaseModel):
    """Configuration for the collected data, containing the data pattern, the background pattern and the bkg scaling
    parameter."""

    data: Pattern = Field(description="The data pattern.")
    bkg: Optional[Pattern] = Field(default=None, description="The background pattern.")
    bkg_scaling: float = Field(
        default=1.0, description="The scaling factor for the background pattern."
    )


class Result(BaseModel):
    calculation_config: CalculationConfig = Field(
        description="The configuration used for the calculation."
    )
    sq: Optional[Pattern] = Field(
        default=None, description="The calculated structure factor S(q)."
    )
    fr: Optional[Pattern] = Field(
        default=None, description="The calculated pair distribution function F(r)."
    )
    gr: Optional[Pattern] = Field(default=None, description="The calculated g(r).")



import sys
import os

from .pattern import Pattern

__version__ = "2.2.1"


def _module_path():
    return os.path.dirname(__file__)


from .utility import *
from .optimization import *
from .soller_correction import *
from .normalization import *
from .transform import *


def _we_are_frozen():
    # All the modules are built-in to the interpreter, e.g., by py2exe
    return hasattr(sys, "frozen")


from enum import Enum


class NormalizationMethod(str, Enum):
    """
    Enum class for the different methods to perform an intensity normalization.
    """

    INTEGRAL = "integral"
    FIT = "fit"


class FourierTransformMethod(str, Enum):
    """
    Enum class for the different methods to perform a Fourier transform.
    """

    FFT = "fft"
    INTEGRAL = "integral"


class ScatteringFactorSource(str, Enum):
    """
    Enum class for the different sources of the scattering factors.
    """

    HAJDU = "hajdu"
    BROWN_HUBBELL = "brown_hubbell"


class NormalizationFitScaling(str, Enum):
    """
    Enum class for the different scaling methods for the fit.
    """

    LINEAR = "linear"
    QUADRATIC = "quadratic"


class ExtrapolationMethod(str, Enum):
    """
    Enum class for the different extrapolation methods of the S(Q) to
    S(0)
    """

    STEP = "step"
    LINEAR = "linear"
    POLY = "poly"
    SPLINE = "spline"




"""
Converts the incoherent scattering data from the individual files to a single csv file. Also converts the s values to Q
values.
"""
import os
import numpy as np
import pandas as pd

files = os.listdir('incoherent')

result = pd.DataFrame()
# hubbell data is saved in s --> Q = 4 * pi * s
result['q'] = np.loadtxt(os.path.join('incoherent', 'Ac.dat'), skiprows=1)[:, 0] * np.pi * 4

for file in files:
    name, ext = os.path.splitext(file)
    if ext == '.dat':
        data = np.loadtxt(os.path.join('incoherent', file), skiprows=1)
        result[name] = data[:, 2]

result.to_csv('incoherent_scattering_intensities.csv', index=False)



# -*- coding: utf-8 -*-

import numpy as np

from .utility import calculate_weighting_factor
from .transform import calculate_fr
from .pattern import Pattern


def i_q_peak(q, n, position, sigma, composition, element_1, element_2):
    """
    Calculates the contribution of one element 1 - element 2 peak in real space to i(Q) (remember i(Q)=S(Q)-1). We
    assume a gaussian broadening. The math is explained in the paper about NXFit (Pickup et al. 2014, J. Appl. Cryst.
    47, 1790-1796).

    :param q: Q value or numpy array with a unit of A^-1
    :param n: coordination number of element 2 to element 1
    :param position: average distance between the two elements
    :param sigma: measure for broadness of distances distribution
    :param composition: composition: dictionary with elements as key and abundances as relative numbers
    :param element_1: string giving element 1
    :param element_2: string giving element 1
    """
    num_atoms = sum([val for _, val in composition.items()])
    c_2 = composition[element_2] / num_atoms
    w = calculate_weighting_factor(composition, element_1, element_2, q)
    return n * w / c_2 * np.sin(q * position) / (q * position) * np.exp(-q ** 2 * sigma ** 2 / 2)


def t_r_peak(r, n, position, sigma, composition, element_1, element_2, q, use_modification_fcn=False, method='fft'):
    """
    Calculates the contribution of one element 1 - element 2 peak in real space to t(r). We assume a gaussian
    broadening. The math is explained in the paper about NXFit (Pickup et al. 2014, J. Appl. Cryst. 47, 1790-1796).

    The function will first calculate the peak in i(Q) with the appropriate weighting factor and then fourier transform
    it into real space.

    :param r: numpy array giving the r-values for which the peak will be calculated
    :param n: coordination number of element 2 to element 1
    :param position: average distance between the two elements
    :param sigma: measure for broadness of distances distribution
    :param composition: composition: dictionary with elements as key and abundances as relative numbers
    :param element_1: string giving element 1
    :param element_2: string giving element 1
    :param q: numpy array giving the q-values for which the peak will be calculated in q-space, should correspond to
                  the same values as the experimental data.
                  WARNING: check whether it works correctly, when your q values are not extended to close 0 A^{-1}.
                  WARNING: q-array should not contain 0, since this will cause a division by zero and the calculation
                  will fail.
    :param use_modification_fcn: boolean flag whether to use the Lorch modification function, during the fourier
                                 transformation.
    :param method: determines the method used for calculating fr, possible values are:
                    - 'integral' solves the Fourier integral, by calculating the integral
                    - 'fft' solves the Fourier integral by using fast fourier transformation
    """
    q_peak = i_q_peak(q, n, position, sigma, composition, element_1, element_2)
    return calculate_fr(Pattern(q, q_peak + 1), r, use_modification_fcn=use_modification_fcn, method=method).y



# -*- coding: utf-8 -*-

import numpy as np


class SollerCorrection(object):
    def __init__(self, two_theta, max_thickness, inner_radius=62, outer_radius=210,
                 inner_width=0.05, outer_width=0.2, inner_length=8, outer_length=6):
        """
        This class handles the calculation of the intensity correction when using soller slits. Upon initialization it
        creates a lookup table for the dispersion angles for each two theta angle and thickness of the sample.

        Corrections are then calculated

        :param two_theta: angles for which the correction needs to be done (in Degree)
        :param max_thickness: maximum sample thickness for which the lookup table (in mm)
        :param inner_radius: radius where the inner slits start (in mm)
        :param outer_radius: radius where the outer slits start (in mm)
        :param inner_width: width of the inner slits (in mm)
        :param outer_width: width of the outer slits (in mm)
        :param inner_length: length of the slit blades (in mm)
        :param outer_length: length of the slit blades (in mm)
        """

        self._two_theta = two_theta / 180. * np.pi
        self._max_thickness = max_thickness
        self._inner_radius = inner_radius
        self._outer_radius = outer_radius
        self._inner_width = inner_width
        self._outer_width = outer_width
        self._inner_length = inner_length
        self._outer_length = outer_length

        self.dispersion_angle_map = self.calculate_dispersion_angle_map()

    def calculate_dispersion_angle_map(self):
        """
        Creates a lookup table of dispersion angles for each two theta value and distance from the center of the
        soller slit rotation center
        :return: a map of the dispersion anges, out.X = two theta array, out.Y = distance array, out.data = dispersion
        angle
        """
        p_x = np.arange(-self._max_thickness * 0.5, self._max_thickness * 0.5, 0.001)
        p_y = np.zeros(p_x.shape)
        p = np.array([p_x, p_y])
        phi_array = []
        for two_theta_value in self._two_theta:
            # calculate fix points for the ther outer parts of the slits
            q1_1, q1_2 = calculate_rectangular_side_points(self._inner_radius + self._inner_length,
                                                           two_theta_value, self._inner_width)

            q2_1, q2_2 = calculate_rectangular_side_points(self._outer_radius + self._outer_length,
                                                           two_theta_value, self._outer_width)

            # calculate fix points for the inner parts of the slits
            s1_1, s1_2 = calculate_rectangular_side_points(self._inner_radius, two_theta_value, self._inner_width)

            # calculate the angles to the outer points of the slits
            phi1 = calculate_angles(q1_1, q1_2, p)
            phi2 = calculate_angles(q2_1, q2_2, p)
            phi3 = calculate_angles(q2_1, q1_2, p)
            phi4 = calculate_angles(q1_1, q2_2, p)

            # take the smallest angle for each point
            phi = np.where(phi1 < phi2, phi1, phi2)
            phi = np.where(phi3 < phi, phi3, phi)
            phi = np.where(phi4 < phi, phi4, phi)

            # getting geometry
            intercept_s1_2_q1_1 = calculate_x_axis_intercept(s1_2, q1_1)
            intercept_s1_2_q2_1 = calculate_x_axis_intercept(s1_2, q2_1)
            intercept_q1_2_q2_2 = calculate_x_axis_intercept(q1_2, q2_2)

            intercept_s1_1_q2_2 = calculate_x_axis_intercept(s1_1, q2_2)
            intercept_q1_1_q2_1 = calculate_x_axis_intercept(q1_1, q2_1)

            pos_cutoff = 0 if intercept_q1_2_q2_2 < 0 else intercept_q1_2_q2_2
            neg_cutoff = 0 if intercept_q1_1_q2_1 > 0 else intercept_q1_1_q2_1

            #####################
            # correcting for positive side:

            intermediate_region_ind = np.logical_and(p_x > pos_cutoff, p_x < intercept_s1_2_q1_1)
            points_in_intermediate_region = p[:, intermediate_region_ind]

            if np.sum(intermediate_region_ind):
                phi1 = calculate_angles(s1_2, q2_1, points_in_intermediate_region)
                phi2 = calculate_angles(q2_2, q2_1, points_in_intermediate_region)
                phi[intermediate_region_ind] = np.where(phi1 < phi2, phi1, phi2)

            # cut the angle
            phi[p[0] > intercept_s1_2_q1_1] = 0
            phi[p[0] > intercept_s1_2_q2_1] = 0

            ########################
            # correcting for negative side:

            intermediate_region_ind = np.logical_and(p_x < neg_cutoff, p_x > intercept_s1_1_q2_2)
            points_in_intermediate_region = p[:, intermediate_region_ind]

            if np.sum(intermediate_region_ind):
                phi1 = calculate_angles(s1_1, q2_2, points_in_intermediate_region)
                phi2 = calculate_angles(q2_1, q2_2, points_in_intermediate_region)
                phi[intermediate_region_ind] = np.where(phi1 < phi2, phi1, phi2)

            intercept_s1_1_q2_2 = calculate_x_axis_intercept(s1_1, q2_2)
            phi[p[0] < intercept_s1_1_q2_2] = 0

            phi_array.append(phi)

        # create the real grid
        two_theta_array_deg = self._two_theta / np.pi * 180
        X, Y = np.meshgrid(two_theta_array_deg, p[0])
        phi_array = np.array(phi_array).transpose()

        return Map(X, Y, phi_array)

    def transfer_function_from_region(self, d1, d2):
        """
        Calculates the transfer function for a sample region within d1 and d2
        :param d1: lower bound of the sample region
        :param d2: upper bound of the sample region
        :return: transfer function with same dimensions as two_theta
        """
        distance = self.dispersion_angle_map.Y[:, 0]
        region_ind = np.logical_and(distance > d1, distance < d2)
        transfer_function = 1. / np.sum(self.dispersion_angle_map.data[region_ind, :], 0)
        transfer_function = transfer_function / np.min(transfer_function)
        return transfer_function

    def transfer_function_sample(self, sample_thickness, shift=0):
        """
        Calculates the transfer function for a specific sample thickness, assuming the sample is centered, to the
        rotation center of the soller slit
        :param sample_thickness: sample thickness in mm
        :param shift: shift of the sample relative to rotation center of the soller slit in beam direction (x)
        :return: transfer function with same dimensions as two_theta
        """
        return self.transfer_function_from_region(-sample_thickness * 0.5 + shift,
                                                  +sample_thickness * 0.5 + shift)

    def transfer_function_dac(self, sample_thickness, initial_thickness):
        """
        Calculates two transfer function specific to diamond anvil cell (DAC) correction. It calculates the transfer
        function for the sample and also for the Compton scattering of diamonds which came into the diffraction volume
        due to the compression.
        :param sample_thickness: current sample thickness in mm
        :param initial_thickness: initial sample chamber thickness when background was measured, in mm
        :return: tuple of (sample transfer function, diamond transfer function), both with same dimensions as two_theta
        """
        sample_transfer_function = self.transfer_function_sample(sample_thickness)
        d1 = sample_thickness * 0.5
        d2 = initial_thickness * 0.5
        diamond_transfer_function = self.transfer_function_from_region(d1, d2)
        diamond_transfer_function += self.transfer_function_from_region(-d2, -d1)
        diamond_transfer_function /= 2
        return sample_transfer_function, diamond_transfer_function
    
class SollerCorrectionGui(SollerCorrection):
    def __init__(self,  q, wavelength, max_thickness, inner_radius=62, outer_radius=210,
                 inner_width=0.05, outer_width=0.2, inner_length=8, outer_length=6):

        two_theta = np.arcsin(q * wavelength / (4 * np.pi)) * 360 / np.pi
        super(SollerCorrectionGui, self).__init__(two_theta, max_thickness, inner_radius, outer_radius,
                 inner_width, outer_width, inner_length, outer_length)
        self.wavelength = wavelength
        self.q = q


# Utility functions and classes
class Map(object):
    def __init__(self, X, Y, data):
        self.X = X
        self.Y = Y
        self.data = data


def vector_length(vec):
    return np.sqrt(np.sum(vec ** 2))


def calculate_angles(point1, point2, p):
    """
    calculates the angle between vectors going from the two points (point1, point2) to a central point p using
    the dot product.
    """
    return np.arccos(((point1[0] - p[0]) * (point2[0] - p[0]) + (point1[1] - p[1]) * (point2[1] - p[1])) /
                     (np.sqrt((point1[0] - p[0]) ** 2 + (point1[1] - p[1]) ** 2) * np.sqrt(
                         (point2[0] - p[0]) ** 2 + (point2[1] - p[1]) ** 2)))


def calculate_x_axis_intercept(p1, p2):
    """
    obtains the x-axis intercept of a line defined by two points.
    """
    m = (p2[1] - p1[1]) / (p2[0] - p1[0])
    c = p2[1] - p2[0] * m
    return -c / m


def calculate_y_axis_intercept(p1, p2):
    """
    obtains the y-axis intercept of a line defined by two points.
    """
    m = (p2[1] - p1[1]) / (p2[0] - p1[0])
    return m * (-p2[0]) + p2[1]


def calculate_rectangular_side_points(radius, angle, width):
    """
    calculates the points which are rectangularly width/2 away at a certain angle, radius combination
    :return: 2 points with (x, y) coordinates
    """
    p1 = np.array([radius * np.cos(angle) - 0.5 * width * np.sin(angle),
                   radius * np.sin(angle) + 0.5 * width * np.cos(angle)])

    p2 = np.array([radius * np.cos(angle) + 0.5 * width * np.sin(angle),
                   radius * np.sin(angle) - 0.5 * width * np.cos(angle)])

    return p1, p2



# -*- coding: utf-8 -*-
from __future__ import annotations
import os
from typing_extensions import Annotated
from typing import Union, Optional
from pydantic import PlainSerializer, PlainValidator, BaseModel
import numpy as np
import base64
import gzip

from dataclasses import dataclass

import io
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter1d


@dataclass
class Pattern:
    """
    A Pattern is a set of x and y values.
    It can be loaded from a file or created from scratch and can be modified by
    different methods.
    It builds the basis for all calculations in glassure.

    :param x: x values of the pattern
    :param y: y values of the pattern
    :param name: name of the pattern
    """

    x: Optional[PydanticNpArray] = None
    y: Optional[PydanticNpArray] = None
    name: str = ""

    def __post_init__(self):
        if self.y is None and self.x is None:
            self.x = np.linspace(0, 10, 101)
            self.y = np.log(self.x**2) - (self.x * 0.2) ** 2

        if len(self.x) != len(self.y):
            raise ValueError("x and y values must have the same length")

    def load(self, filename: str, skiprows: int = 0):
        """
        Loads a pattern from a file. The file can be either a .xy or a .chi file. The .chi file will be loaded with
        skiprows=4 by default.

        :param filename: path to the file
        :param skiprows: number of rows to skip when loading the data (header)
        """
        try:
            if filename.endswith(".chi"):
                skiprows = 4
            data = np.loadtxt(filename, skiprows=skiprows)
            self.x = data.T[0]
            self.y = data.T[1]
            self.name = os.path.basename(filename).split(".")[:-1][0]

        except ValueError:
            print("Wrong data format for pattern file! - " + filename)
            return -1

    @staticmethod
    def from_file(filename: str, skip_rows: int = 0) -> Pattern | "-1":
        """
        Loads a pattern from a file. The file can be either a .xy or a .chi file. The .chi file will be loaded with
        skiprows=4 by default.

        :param filename: path to the file
        :param skip_rows: number of rows to skip when loading the data (header)
        """
        try:
            if filename.endswith(".chi"):
                skip_rows = 4
            data = np.loadtxt(filename, skiprows=skip_rows)
            x = data.T[0]
            y = data.T[1]
            name = os.path.basename(filename).split(".")[:-1][0]
            return Pattern(x, y, name)

        except ValueError:
            print("Wrong data format for pattern file! - " + filename)
            return -1

    def save(self, filename: str, header: str = ""):
        """
        Saves the Pattern to a two-column xy file.

        :param filename: path to the file
        :param header: header to be written to the file
        """
        data = np.dstack((self.x, self.y))
        np.savetxt(filename, data[0], header=header)

    def smooth(self, amount: float) -> Pattern:
        """
        Smoothing the pattern by applying a gaussian filter. Returns the smoothed pattern.
        :param amount: amount of smoothing to be applied
        """
        return Pattern(self.x, gaussian_filter1d(self.y, amount))

    def rebin(self, bin_size: float) -> Pattern:
        """
        Returns a new pattern, which is a rebinned version of the current one.

        :param bin_size: Size of the bins
        :return: rebinned Pattern
        """
        x, y = self.data
        x_min = np.round(np.min(x) / bin_size) * bin_size
        x_max = np.round(np.max(x) / bin_size) * bin_size
        new_x = np.arange(x_min, x_max + 0.1 * bin_size, bin_size)

        bins = np.hstack((x_min - bin_size * 0.5, new_x + bin_size * 0.5))
        new_y = np.histogram(x, bins, weights=y)[0] / np.histogram(x, bins)[0]

        return Pattern(new_x, new_y)

    @property
    def data(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns the data of the pattern as a tuple of x and y values.

        :return: tuple of x and y values
        """
        return self.x, self.y

    @data.setter
    def data(self, data: tuple[np.ndarray, np.ndarray]):
        """
        Sets the data of the pattern. Also resets the scaling and offset to 1 and 0 respectively.

        :param data: tuple of x and y values
        """
        self.x = data[0]
        self.y = data[1]

    def limit(self, x_min: float, x_max: float) -> Pattern:
        """
        Limits the pattern to a specific x-range. Does not modify inplace but returns a new limited Pattern

        :param x_min: lower limit of the x-range
        :param x_max: upper limit of the x-range
        :return: limited Pattern
        """
        x, y = self.data
        return Pattern(
            x[np.where((x_min < x) & (x < x_max))],
            y[np.where((x_min < x) & (x < x_max))],
        )

    def extend_to(self, x_value: float, y_value: float) -> Pattern:
        """
        Extends the current pattern to a specific x_value by filling it with the y_value. Does not modify inplace but
        returns a new filled Pattern

        :param x_value: Point to which extending the pattern should be smaller than the lowest x-value in the pattern or
        vice versa
        :param y_value: number to fill the pattern with
        :return: extended Pattern
        """
        x_step = np.mean(np.diff(self.x))
        x_min = np.min(self.x)
        x_max = np.max(self.x)
        if x_value < x_min:
            x_fill = np.arange(x_min - x_step, x_value - x_step * 0.5, -x_step)[::-1]
            y_fill = np.zeros(x_fill.shape)
            y_fill.fill(y_value)

            new_x = np.concatenate((x_fill, self.x))
            new_y = np.concatenate((y_fill, self.y))
        elif x_value > x_max:
            x_fill = np.arange(x_max + x_step, x_value + x_step * 0.5, x_step)
            y_fill = np.zeros(x_fill.shape)
            y_fill.fill(y_value)

            new_x = np.concatenate((self.x, x_fill))
            new_y = np.concatenate((self.y, y_fill))
        else:
            return self

        return Pattern(new_x, new_y)

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the pattern which can be used to save the pattern to a json file.

        :return: dictionary representation of the pattern
        """
        return {
            "name": self.name,
            "x": self.x.tolist(),
            "y": self.y.tolist(),
        }

    @staticmethod
    def from_dict(json_dict: dict) -> Pattern:
        """
        Creates a new Pattern from a dictionary representation of a Pattern.

        :param json_dict: dictionary representation of a Pattern
        :return: new Pattern
        """
        return Pattern(
            np.array(json_dict["x"]), np.array(json_dict["y"]), json_dict["name"]
        )

    ###########################################################
    # Operators:

    def __sub__(self, other: Union[float, Pattern]) -> Pattern:
        """
        Subtracts a float or another pattern from the current one. If the other pattern
        has a different shape, the subtraction will be done on the overlapping
        x-values and the background will be interpolated. If there is no
        overlapping between the two patterns, a BkgNotInRangeError will be
        raised.

        :param other: Pattern to be subtracted
        :return: new Pattern
        """
        return self.__add__(-1 * other)

    def __rsub__(self, other: Union[float, Pattern]) -> Pattern:
        """
        Subtracts the pattern from a float or another pattern. If the other
        pattern has a different shape, the subtraction will be done on the
        overlapping x-values and the background will be interpolated. If there
        is no overlapping between the two patterns, a BkgNotInRangeError will
        be raised.

        :param other: float or Pattern to be subtracted
        :return: new Pattern
        """
        return -1 * self.__sub__(other)

    def __add__(self, other: Union[float, Pattern]) -> Pattern:
        """
        Adds the other pattern to the current one. If the other pattern
        has a different shape, the addition will be done on the overlapping
        x-values and the y-values of the other pattern will be interpolated.
        If there is no overlapping between the two patterns, a BkgNotInRangeror
        will be raised.

        :param other: Pattern to be added
        :return: new Pattern
        """
        orig_x, orig_y = self.data

        if isinstance(other, (int, float)):
            return Pattern(orig_x, orig_y + other)

        other_x, other_y = other.data

        if orig_x.shape != other_x.shape:
            # the background will be interpolated
            other_fcn = interp1d(other_x, other_y, kind="linear")

            # find overlapping x and y values:
            ind = np.where((orig_x <= np.max(other_x)) & (orig_x >= np.min(other_x)))
            x = orig_x[ind]
            y = orig_y[ind]

            if len(x) == 0:
                # if there is no overlapping between background and pattern, raise an error
                raise BkgNotInRangeError(self.name)
            return Pattern(x, y + other_fcn(x))
        else:
            return Pattern(orig_x, orig_y + other_y)

    def __radd__(self, other: Union[float, Pattern]) -> Pattern:
        """
        Adds the other pattern to the current one. If the other pattern
        has a different shape, the addition will be done on the overlapping
        x-values and the y-values of the other pattern will be interpolated.
        If there is no overlapping between the two patterns, a BkgNotInRangeror
        will be raised.

        :param other: Pattern to be added
        :return: new Pattern
        """
        return self.__add__(other)

    def __rmul__(self, other: float) -> Pattern:
        """
        Multiplies the pattern with a scalar.

        :param other: scalar to multiply with
        :return: new Pattern
        """
        orig_x, orig_y = self.data
        return Pattern(np.copy(orig_x), np.copy(orig_y) * other)

    def __mul__(self, other: float) -> Pattern:
        """
        Multiplies the pattern with a scalar.

        :param other: scalar to multiply with
        :return: new Pattern
        """
        return self.__rmul__(other)

    def __eq__(self, other: Pattern) -> bool:
        """
        Checks if two patterns are equal. Two patterns are equal if their data
        is equal.

        :param other: Pattern to compare with
        :return: True if equal, False otherwise
        """
        if not isinstance(other, Pattern):
            return False
        if np.array_equal(self.data, other.data):
            return True
        return False


class BkgNotInRangeError(Exception):
    def __init__(self, pattern_name: str):
        self.pattern_name = pattern_name

    def __str__(self):
        return (
            "The background range does not overlap with the Pattern range for "
            + self.pattern_name
        )


def validate(value):
    if isinstance(value, list):
        return np.array(value)
    if isinstance(value, str):
        try:
            v = base64.b64decode(value)
            v = gzip.decompress(v)
            numpy_array = np.load(io.BytesIO(v), allow_pickle=False)
            return numpy_array
        except Exception as e:
            raise ValueError(f"Could not decode numpy array: {e}")
    if isinstance(value, np.ndarray):
        return value
    raise TypeError(f"Invalid type for numpy array: {type(value)}")


def serialize(value):
    if isinstance(value, np.ndarray):
        # Save numpy array to compressed bytyes
        with io.BytesIO() as buffer:
            np.save(buffer, value, allow_pickle=False)
            binary_data = buffer.getvalue()
        compressed_data = gzip.compress(binary_data)
        base_64_data = base64.b64encode(compressed_data).decode("utf-8")
        return base_64_data
    raise TypeError(f"Invalid type for numpy array: {type(value)}")


PydanticNpArray = Annotated[
    np.ndarray, PlainValidator(validate), PlainSerializer(serialize)
]



# -*- coding: utf-8 -*-

from copy import deepcopy
from typing import Optional

import numpy as np
from scipy.integrate import simpson
from lmfit import Parameters, minimize

from . import Pattern
from .transform import calculate_fr, calculate_gr, calculate_sq_from_fr

__all__ = [
    "optimize_sq",
    "optimize_density",
]


def optimize_sq(
    sq_pattern: Pattern,
    r_cutoff: float,
    iterations: int,
    atomic_density: float,
    r_step: float = 0.01,
    use_modification_fcn: bool = False,
    attenuation_factor: float = 1,
    fcn_callback=None,
    callback_period: int = 2,
    fourier_transform_method: str = "fft",
):
    """
    Performs an optimization of the structure factor based on an r_cutoff value as described in Eggert et al. 2002 PRB,
    65, 174105. This basically does back and forward transforms between S(Q) and f(r) until the region below the
    r_cutoff value is a flat line without any oscillations.

    :param sq_pattern:
        original S(Q)
    :param r_cutoff:
        cutoff value below which there is no signal expected (below the first peak in g(r))
    :param iterations:
        number of back and forward transforms
    :param atomic_density:
        density in atoms/A^3
    :param r_step:
        step size for the r-axis, default is 0.01. Use smaller values for better accuracy (especially if needed for
         fft)
    :param use_modification_fcn:
        Whether to use the Lorch modification function during the Fourier transform.
        Warning: When using the Lorch modification function, usually more iterations are needed to get to the
        wanted result.
    :param attenuation_factor:
        Sometimes the initial change during back and forward transformations results in a runaway, by setting the
        attenuation factor reduces the amount of change during each iteration.
    :param fcn_callback:
        Function which will be called at an iteration period defined by the callback_period parameter.
        The function should take three arguments: sq_pattern, fr_pattern and gr_pattern.
        Additionally, the function should return a boolean value, where True continues the optimization and False will
        stop the optimization.
    :param callback_period:
        determines how frequently the fcn_callback will be called.
    :param fourier_transform_method:
        determines which method will be used for the Fourier transform. Possible values are 'fft' and 'integral'

    :return:
        optimized S(Q) pattern
    """
    r = np.arange(0, r_cutoff, r_step)
    sq_pattern = deepcopy(sq_pattern)
    for iteration in range(iterations):
        fr_pattern = calculate_fr(
            sq_pattern, r, use_modification_fcn, method=fourier_transform_method
        )
        q, sq_int = sq_pattern.data
        r, fr_int = fr_pattern.data

        delta_fr = fr_int + 4 * np.pi * r * atomic_density

        if fourier_transform_method == "fft":
            sq_trans_fft = (
                calculate_sq_from_fr(Pattern(r, delta_fr), sq_pattern.x, method="fft")
                - 1
            )
            iq = sq_trans_fft.y
        else:
            in_integral = np.array(np.sin(np.outer(q.T, r))) * delta_fr
            iq = np.trapz(in_integral, r) / q

        sq_pattern = sq_pattern * (1 - iq / attenuation_factor)

        if fcn_callback is not None and iteration % callback_period == 0:
            fr_pattern = calculate_fr(
                sq_pattern,
                use_modification_fcn=use_modification_fcn,
                method=fourier_transform_method,
            )
            gr_pattern = calculate_gr(fr_pattern, atomic_density)
            fcn_callback(sq_pattern, fr_pattern, gr_pattern)
    return sq_pattern


from .calc import calculate_pdf
from .configuration import DataConfig, CalculationConfig
from .methods import ExtrapolationMethod


def optimize_density(
    data_config: DataConfig,
    calculation_config: CalculationConfig,
    method: str = "fr",
    min_range: Optional[tuple[float, float]] = None,
    vary_bkg_scaling: bool = True,
    bkg_limits: tuple[float, float] = (0.9, 1.1),
    optimization_method: str = "lsq",
) -> tuple[float, float, float, float]:
    """
    Optimizes the density of the sample using the g(r), f(r) or S(Q) (chosen by the method parameter).
    The density in the SampleConfig of the DataConfig is taking as starting parameter

    For method='gr' or method='fr' the optimization is based on the g(r) or f(r) function, and the density is
    optimized to minimize the low g(r) or f(r) region to be close to zero. The Lorch modification function will be 
    applied before calculating the chi square of the low r region if it is applied in the calculation configuration. 
    The general procedure is explained in Eggert et al. 2002 PRB, 65, 174105.

    For method='sq' the optimization is based on the low Q part of the S(Q) function, and the density is optimized
    to minimize the difference between the original S(Q) function without any optimization and the optimized S(Q)
    function. The configuration should have extrapolation enabled for this to work best.
    For polyatomic systems, finding the density using this procedure is much less susceptible to the Q_max value of
    the S(Q) than the g(r) based optimization. However, density is not exactly the same for both methods and the
    method needs to be verified further. (Please us the method='sq' with caution.)

    The best for both types is to have a reference density to compare it to. Based on this then further calculations
    of e.g. high pressure or high temperature densities can be performed.

    For this procedure to work best, the S(Q) optimization should be enabled in the calculation configuration. The
    chosen parameters are then used in the find density function.

    example usage:
    ```
    from glassure.calc import create_calculate_pdf_configs
    from glassure.optimization import optimize_density

    data_config, calculation_config = create_calculate_pdf_configs(data, composition, density, background)
    calculation_config.transform.q_min = 1
    calculation_config.transform.q_max = 16
    calculation_config.transform.extrapolation.method = ExtrapolationMethod.LINEAR
    calculation_config.optimize = OptimizeConfig(r_cutoff=1.4)

    density, density_error, bkg_scaling, bkg_error = optimize_density(data_config, calculation_config, method='gr', range=(0.1, 1.2))
    ```

    :param data_config:
        Data configuration
    :param calculation_config:
        Calculation configuration
    :param method:
        Method to use for the optimization. Possible values are 'gr', 'fr' and 'sq'.
    :param min_range:
        x range of the data to use for the minimization to find the density. For method='gr' and 'fr this is the
        r-range of the g(r)/f(r) function to minimize. For method='sq' this is the Q-range of the S(Q) function to
        minimize the difference between the original and optimized S(Q) function. Default is None which means that
        the range is (0, calculation_config.optimize.r_cutoff) for method='gr' and 'fr' and
        (0, calculation_config.transform.q_max) for method='sq'.
    :param vary_bkg_scaling:
        Whether to vary the background scaling during the optimization. Default is True.
    :param bkg_limits:
        relative limits for the background scaling. The background scaling is optimized to be within these limits.
        Default is (0.9, 1.1) which means that the background scaling is optimized to be within 10% of the starting
        value.
    :param optimization_method:
        Method to use for the optimization. Possible values are 'nelder' and 'lsq'.

    :return:
        a tuple with four values:
        - the density,  its error value, the background scaling and the error value
        whereby the error value is the standard error of the fit parameter for optimization_method ='lsq' and the
        sum of the squared residuals for optimization_method='nelder'.
    """

    if calculation_config.optimize is None and min_range is None and not method == "sq":
        raise ValueError(
            "For optimizing density using 'gr' or 'fr' the calculation configuration needs to have the "
            "optimize configuration or the min_range parameter needs to be set."
        )

    params = Parameters()
    params.add("density", value=calculation_config.sample.density, min=0.0, max=100)
    params.add(
        "bkg_scaling",
        value=data_config.bkg_scaling,
        min=data_config.bkg_scaling * bkg_limits[0],
        max=data_config.bkg_scaling * bkg_limits[1],
        vary=vary_bkg_scaling,
    )

    optim_config = calculation_config.model_copy(deep=True)

    if method == "sq":
        reference_config = calculation_config.model_copy(deep=True)
        reference_config.optimize = None
        reference_result = calculate_pdf(data_config, reference_config)
        if min_range is None:
            min_range = (0, reference_config.transform.q_max)

    elif method == "gr" or method == "fr":
        if min_range is None:
            min_range = (0, optim_config.optimize.r_cutoff)

    def fcn(params):
        density = params["density"].value
        bkg_scaling = params["bkg_scaling"].value
        optim_config.sample.density = density
        data_config.bkg_scaling = bkg_scaling
        result = calculate_pdf(data_config, optim_config)

        if method == "gr":
            r, gr = result.gr.limit(*min_range).data
            residual = gr * (r[1] - r[0])
        elif method == "fr":
            atomic_density = optim_config.sample.atomic_density
            r, fr = result.fr.limit(*min_range).data
            residual = (fr + 4 * np.pi * r * atomic_density) * (r[1] - r[0])
        elif method == "sq":
            q, sq = result.sq.limit(*min_range).data
            sq_ref = reference_result.sq.limit(*min_range).y
            residual = (sq - sq_ref) * (q[1] - q[0])
        else:
            raise ValueError(
                f"Invalid optimize density method: {method}, only 'gr', 'fr' and 'sq' are supported."
            )
        return residual

    if optimization_method == "nelder":
        res = minimize(
            fcn,
            params,
            method="nelder",
            options={"maxfev": 500, "fatol": 0.0001, "xatol": 0.0001},
        )
        return (
            res.params["density"].value,
            np.sum(res.residual**2),
            res.params["bkg_scaling"].value,
            np.sum(res.residual**2),
        )
    elif optimization_method == "lsq":
        res = minimize(
            fcn,
            params,
            method="least_squares",
        )
        return (
            res.params["density"].value,
            res.params["density"].stderr,
            res.params["bkg_scaling"].value,
            res.params["bkg_scaling"].stderr,
        )
    else:
        raise ValueError(f"Invalid optimization method: {optimization_method}")



# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import UnivariateSpline
from .pattern import Pattern


def calculate_transfer_function(std_pattern: Pattern, sample_pattern: Pattern, smooth_factor=1) -> UnivariateSpline:
    """

    :param std_pattern: the Diffraction pattern how it should look like, should be already background subtracted
    :type std_pattern: Pattern
    :param sample_pattern: the Diffraction pattern of the same sample which needs a transfer function
    :param smooth_factor: Determines the amount of smoothing of the transfer function
    :return:
    """
    transfer_function = std_pattern.y / sample_pattern.y
    transfer_function /= np.min(transfer_function)

    return UnivariateSpline(std_pattern.x, transfer_function, k=3, s=smooth_factor)



# -*- coding: utf-8 -*-
import os
import re
from abc import ABC, abstractmethod

import numpy as np
import scipy
from scipy.interpolate import interp1d
import pandas
from . import _module_path

import xraylib as xr

module_data_path = os.path.join(_module_path(), "data")

atomic_weights = pandas.read_csv(
    os.path.join(module_data_path, "atomic_weights.csv"), index_col=0
)


class ScatteringFactorCalculator(ABC):
    """
    Abstract class for scattering factor calculators.
    """

    @abstractmethod
    def get_coherent_scattering_factor(self, element: str, q):
        raise NotImplementedError

    @abstractmethod
    def get_incoherent_intensity(self, element: str, q):
        raise NotImplementedError

    @property
    def elements(self):
        raise NotImplementedError


class ScatteringFactorCalculatorHajdu(ScatteringFactorCalculator):
    """
    Scattering factor calculator based on the work of Hajdu et al. (Acta Cryst. (1992). A48, 344-352).
    """

    def __init__(self):
        self.coherent_param = pandas.read_csv(
            os.path.join(
                module_data_path, "hajdu", "param_coherent_scattering_factors.csv"
            ),
            index_col=0,
        )
        self.incoherent_param = pandas.read_csv(
            os.path.join(
                module_data_path, "hajdu", "param_incoherent_scattering_intensities.csv"
            ),
            index_col=0,
        )

    def get_coherent_scattering_factor(self, element: str, q):
        """
        Calculates the coherent scattering factor for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: coherent scattering factor array
        """
        if element not in self.coherent_param.index.values:
            raise ElementNotImplementedException(element)
        fs_coh = 0
        s = q / (4 * np.pi)
        for ind in range(1, 5):
            A = self.coherent_param["A" + str(ind)][element]
            B = self.coherent_param["B" + str(ind)][element]
            fs_coh += A * np.exp(-B * s**2)

        C = self.coherent_param["C"][element]
        fs_coh += C
        return fs_coh

    def get_incoherent_intensity(self, element: str, q):
        """
        Calculates the incoherent scattering intensity for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: incoherent scattering intensity array
        """
        fs_coherent = self.get_coherent_scattering_factor(element, q)
        intensity_coherent = fs_coherent**2
        s = q / (4 * np.pi)
        Z = float(self.incoherent_param["Z"][element])
        M = float(self.incoherent_param["M"][element])
        K = float(self.incoherent_param["K"][element])
        L = float(self.incoherent_param["L"][element])
        intensity_incoherent = (Z - intensity_coherent / Z) * (
            1 - M * (np.exp(-K * s) - np.exp(-L * s))
        )
        return intensity_incoherent

    @property
    def elements(self):
        """
        Returns a list of available elements.
        """
        return self.coherent_param.index.values


class ScatteringFactorCalculatorBrownHubbell(ScatteringFactorCalculator):
    """
    Scattering factor calculator based on the work of Brown et al., 2006 and Hubbell et al., 1975.
    """

    def __init__(self):
        self.coherent_params = pandas.read_csv(
            os.path.join(
                module_data_path,
                "brown_hubbell",
                "param_coherent_scattering_factors.csv",
            ),
            index_col=0,
        )

        self.incoherent_intensities = pandas.read_csv(
            os.path.join(
                module_data_path,
                "brown_hubbell",
                "incoherent_scattering_intensities.csv",
            )
        )

    def get_coherent_scattering_factor(self, element: str, q):
        """
        Calculates the coherent scattering factor for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: coherent scattering factor array
        """
        if element not in self.coherent_params.index.values:
            raise ElementNotImplementedException(element)
        fs_coh = 0
        s = q / (4 * np.pi)
        for ind in range(1, 5):
            A = self.coherent_params["a" + str(ind)][element]
            B = self.coherent_params["b" + str(ind)][element]
            fs_coh += A * np.exp(-B * s**2)

        C = self.coherent_params["c"][element]
        fs_coh += C
        return fs_coh

    def get_incoherent_intensity(self, element: str, q):
        """
        Calculates the incoherent scattering intensity for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: incoherent scattering intensity array
        """
        # use regular expression to find element string of input
        element = re.findall("[A-zA-Z]*", element)[0]
        if element not in self.incoherent_intensities.keys():
            raise ElementNotImplementedException(element)
        interp = scipy.interpolate.interp1d(
            self.incoherent_intensities["q"],
            self.incoherent_intensities[element],
            kind="cubic",
        )
        return interp(q)

    @property
    def elements(self):
        """
        Returns a list of available elements.
        """
        return self.coherent_params.index.values


class ScatteringFactorCalculatorXraylib(ScatteringFactorCalculator):
    """
    Scattering factor calculator using xraylib library. See https://github.com/tschoonj/xraylib/wiki
    for more details.
    """

    def __init__(self):
        """
        Initializes the xraylib scattering factor calculator.
        """

        self._elements = [xr.AtomicNumberToSymbol(i) for i in range(1, 108)]

    def get_coherent_scattering_factor(self, element: str, q):
        """
        Calculates the coherent scattering factor for a given element and q values.

        :param element: Element symbol
        :param q: q array
        :return: coherent scattering factor array
        """
        q_xr = q / (4 * np.pi)
        element_number = xr.SymbolToAtomicNumber(element)
        return np.array([xr.FF_Rayl(element_number, v) for v in q_xr])

    def get_incoherent_intensity(self, element: str, q):
        """
        Calculates the incoherent scattering intensity for a given element and q values,
        using SF_Compt for q_xr >= 0.001, and cubic interpolation for q_xr < 0.001.

        :param element: Element symbol
        :param q: 1D numpy array of q values
        :return: numpy array of incoherent scattering intensity
        """
        element_number = xr.SymbolToAtomicNumber(element)
        q_xr = q / (4 * np.pi)

        # Split into high and low q_xr regions
        low_mask = q_xr < 0.001
        high_mask = ~low_mask

        # Initialize result array
        values = np.zeros_like(q_xr)

        # Compute SF_Compt for q_xr >= 0.001
        q_xr_high = q_xr[high_mask]
        sf_high = np.array([xr.SF_Compt(element_number, v) for v in q_xr_high])
        values[high_mask] = sf_high

        # Interpolate for q_xr < 0.001
        if np.any(low_mask) and q_xr_high.size >= 4:  # at least 4 points for cubic
            n_interp = min(3, q_xr_high.size)
            left_interp_x = np.arange(0, -1, -(q_xr[1] - q_xr[0]))
            left_interp_y = np.zeros_like(left_interp_x)
            interp_q = np.concatenate((left_interp_x, q_xr_high[:n_interp]))
            interp_sf = np.concatenate((left_interp_y, sf_high[:n_interp]))

            f_interp = interp1d(
                interp_q, interp_sf, kind="cubic", bounds_error=False, fill_value=0.0
            )

            values[low_mask] = f_interp(q_xr[low_mask])
        else:
            # If not enough points for interpolation, set to zero
            values[low_mask] = 0.0

        return values

    @property
    def elements(self):
        """
        Returns a list of available elements.
        """
        return self._elements


calculators = {
    "hajdu": ScatteringFactorCalculatorHajdu(),
    "brown_hubbell": ScatteringFactorCalculatorBrownHubbell(),
    "xraylib": ScatteringFactorCalculatorXraylib(),
}

sources = calculators.keys()


def get_calculator(source: str) -> ScatteringFactorCalculator:
    """
    Returns the calculator for a given source. Possible sources are 'hajdu', 'brown_hubbell' and 'xraylib'.
    """
    if source not in calculators.keys():
        raise SourceNotImplementedException(source)
    return calculators[source]


def get_available_elements(source: str) -> list[str]:
    """
    Returns a list of available elements for a given source. Possible sources are 'hajdu', 'brown_hubbell' and
    'xraylib'.
    """
    return get_calculator(source).elements


def calculate_coherent_scattering_factor(
    element: str, q: np.array, source: str = "hajdu"
) -> np.array:
    """
    Calculates the coherent scattering factor for a given element and q values.

    :param element: Element symbol
    :param q: q array in A^-1
    :param source: Source of the scattering factors. Possible sources are 'hajdu', 'brown_hubbell' and 'xraylib'.
    :return: coherent scattering factor array
    """
    return get_calculator(source).get_coherent_scattering_factor(element, q)


def calculate_incoherent_scattered_intensity(
    element: str, q: np.array, source: str = "hajdu"
) -> np.array:
    """
    Calculates the incoherent scattering intensity for a given element and q values.

    :param element: Element symbol
    :param q: q array
    :param source: Source of the scattering factors. Possible sources are 'hajdu', 'brown_hubbell' and 'xraylib'.
    :return: incoherent scattering intensity array
    """
    return get_calculator(source).get_incoherent_intensity(element, q)


class ElementNotImplementedException(Exception):
    def __init__(self, element):
        self.element = element

    def __str__(self):
        return repr("Element " + self.element + " not known or available.")


class SourceNotImplementedException(Exception):
    def __init__(self, source):
        self.source = source

    def __str__(self):
        return repr("Source " + self.source + " not known or available.")
